;;; guile-assimp, foreign interface to libassimp
;;; Copyright (C) 2014 by Javier Sancho Fernandez <jsf at jsancho dot org>
;;;
;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (assimp assimp)
  #:use-module (assimp low-level)
  #:use-module (assimp low-level color)
  #:use-module (assimp low-level material)
  #:use-module (assimp low-level mesh)
  #:use-module (assimp low-level scene)
  #:use-module (assimp low-level vector)
  #:use-module (system foreign))

(define libassimp (dynamic-link "libassimp"))

(define aiImportFile
  (pointer->procedure '*
		      (dynamic-func "aiImportFile" libassimp)
		      (list '* unsigned-int)))


;;; Scenes

(define-conversion-type parse-aiScene -> scene
  (flags (field 'mFlags))
  (root-node (wrap (field 'mRootNode) wrap-node))
  (meshes (wrap (array (field 'mNumMeshes) (field 'mMeshes)) wrap-mesh))
  (materials (wrap (array (field 'mNumMaterials) (field 'mMaterials)) wrap-material))
  (animations (array (field 'mNumAnimations) (field 'mAnimations)))
  (textures (array (field 'mNumTextures) (field 'mTextures)))
  (lights (array (field 'mNumLights) (field 'mLights)))
  (cameras (array (field 'mNumCameras) (field 'mCameras))))

(define (load-scene filename flags)
  (wrap-scene
   (aiImportFile (string->pointer filename)
 		 flags)))

(export load-scene
	scene?
	scene-contents
	scene-flags
	scene-root-node
	scene-meshes
	scene-materials
	scene-animations
	scene-textures
	scene-lights
	scene-cameras)


;;; Nodes

(define-conversion-type parse-aiNode -> node
  (name (sized-string (field 'mName)))
  (transformation (field 'mTransformation))
  (parent (wrap (field 'mParent) wrap-node))
  (children (wrap (array (field 'mNumChildren) (field 'mChildren)) wrap-node))
  (meshes (array (field 'mNumMeshes) (field 'mMeshes))))

(export node?
	node-contents
	node-name
	node-transformation
	node-parent
	node-children
	node-meshes)


;;; Meshes

(define-conversion-type parse-aiMesh -> mesh
  (name (sized-string (field 'mName)))
  (primitive-types (field 'mPrimitiveTypes))
  (vertices (wrap
	     (array (field 'mNumVertices) (field 'mVertices) #:element-proc get-element-address)
	     wrap-vector3d))
  (faces (wrap
	  (array (field 'mNumFaces) (field 'mFaces) #:element-size 8 #:element-proc get-element-address)
	  wrap-face))
  (normals (wrap
	    (array (field 'mNumVertices) (field 'mNormals) #:element-size 12 #:element-proc get-element-address)
	    wrap-vector3d))
  (tangents (wrap
	     (array (field 'mNumVertices) (field 'mTangents) #:element-size 12 #:element-proc get-element-address)
	     wrap-vector3d))
  (bitangents (wrap
	       (array (field 'mNumVertices) (field 'mBitangents) #:element-size 12 #:element-proc get-element-address)
	       wrap-vector3d))
  (colors (map
	   (lambda (c)
	     (wrap
	      (array (field 'mNumVertices) c #:element-size 16 #:element-proc get-element-address)
	      wrap-color4d))
	   (field 'mColors)))
  (texture-coords (map
		   (lambda (tc)
		     (wrap
		      (array (field 'mNumVertices) tc #:element-size 12 #:element-proc get-element-address)
		      wrap-vector3d))
		   (field 'mTextureCoords)))
  (num-uv-components (field 'mNumUVComponents))
  (bones (wrap (array (field 'mNumBones) (field 'mBones)) wrap-bone))
  (material-index (field 'mMaterialIndex)))

(export mesh?
	mesh-contents
	mesh-name
	mesh-primitive-types
	mesh-vertices
	mesh-faces
	mesh-normals
	mesh-tangents
	mesh-bitangents
	mesh-colors
	mesh-texture-coords
	mesh-num-uv-components
	mesh-bones
	mesh-material-index)


;;; Materials

(define-conversion-type parse-aiMaterial -> material
  (properties (array (field 'mNumProperties) (field 'mProperties)))
  (num-allocated (field 'mNumAllocated)))

(export material?
	material-contents
	material-properties
	material-num-allocated)


;;; Faces

(define-conversion-type parse-aiFace -> face
  (indices (array (field 'mNumIndices) (field 'mIndices))))

(export face?
	face-contents
	face-indices)


;;; Vectors

(define-conversion-type parse-aiVector2D -> vector2d
  (x (field 'x))
  (y (field 'y)))

(export vector2d?
	vector2d-contents
	vector2d-x
	vector2d-y)

(define-conversion-type parse-aiVector3D -> vector3d
  (x (field 'x))
  (y (field 'y))
  (z (field 'z)))

(export vector3d?
	vector3d-contents
	vector3d-x
	vector3d-y
	vector3d-z)


;;; Colors

(define-conversion-type parse-aiColor4D -> color4d
  (r (field 'r))
  (g (field 'g))
  (b (field 'b))
  (a (field 'a)))

(export color4d?
	color4d-contents
	color4d-r
	color4d-g
	color4d-b
	color4d-a)


;;; Bones

(define-conversion-type parse-aiBone -> bone
  (name (sized-string (field 'mName)))
  (weights (wrap
	    (array (field 'mNumWeights) (field 'mWeights) #:element-size 8 #:element-proc get-element-address)
	    wrap-vertex-weight))
  (offset-matrix (field 'mOffsetMatrix)))

(export bone?
	bone-contents
	bone-name
	bone-weights
	bone-offset-matrix)


;;; Weights

(define-conversion-type parse-aiVertexWeight -> vertex-weight
  (vertex-id (field 'mVertexId))
  (weight (field 'mWeight)))

(export vertex-weight?
	vertex-weight-contents
	vertex-weight-vertex-id	
	vertex-weight-weight)
